from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import IO4Edge
from ekfsm.devices.utils import retry
from ekfsm.log import ekfsm_logger
from io4edge_client.colorLED import Client, Pb

logger = ekfsm_logger(__name__)


class LEDArray(Device):
    """
    Device class for handling a LED array.
    """

    def __init__(
        self,
        name: str,
        parent: IO4Edge,
        children: list[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        *args,
        **kwargs,
    ):
        logger.debug(
            f"Initializing LEDArray '{name}' with parent device {parent.deviceId}"
        )

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name

        if service_suffix is not None:
            self.service_suffix = service_suffix
            logger.debug(f"Using custom service suffix: {service_suffix}")
        else:
            self.service_suffix = name
            logger.debug(f"Using default service suffix: {name}")

        self.service_addr = f"{parent.deviceId}-{self.service_suffix}"
        logger.info(
            f"LEDArray '{name}' configured with service address: {self.service_addr}"
        )

        try:
            self.client = Client(self.service_addr, connect=False)
            logger.debug(f"LEDArray client created for service: {self.service_addr}")
        except Exception as e:
            logger.error(
                f"Failed to create LEDArray client for {self.service_addr}: {e}"
            )
            raise

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"


class ColorLED(Device):
    """
    Device class for handling a color LED.
    """

    def __init__(
        self,
        name: str,
        parent: "LEDArray",
        children: list[Device] | None = None,
        abort: bool = False,
        channel_id: int = 0,
        *args,
        **kwargs,
    ):
        logger.debug(f"Initializing ColorLED '{name}' on channel {channel_id}")

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name
        self.channel_id = channel_id

        self.client = parent.client
        logger.info(
            f"ColorLED '{name}' initialized on channel {channel_id} with parent LEDArray"
        )

    @retry()
    def describe(self):
        pass

    @retry()
    def get(self) -> tuple[int, int, int, bool]:
        """
        Get color LED state.

        Returns
        -------
            Current color and blink state.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        logger.info(
            "Getting color LED state for '%s' on channel %s", self.name, self.channel_id
        )
        try:
            result = self.client.get(self.channel_id)
            r, g, b, blink = result
            logger.info(
                "ColorLED '%s' state: color=(%s, %s, %s), blink=%s",
                self.name,
                r,
                g,
                b,
                blink,
            )
            return result
        except Exception as e:
            logger.error(
                "Failed to get ColorLED '%s' state on channel %s: %s",
                self.name,
                self.channel_id,
                e,
            )
            raise

    @retry()
    def set(
        self, color: Pb.Color | Pb.RGBColor | tuple[int, int, int] | str, blink: bool
    ) -> None:
        """
        Set the color of the color LED.

        Parameters
        ----------
        color : Pb.Color | Pb.RGBColor | tuple[int, int, int] | str
            Color to set the LED to, given as Pb.Color, Pb.RGBColor, a tuple of
            (red, green, blue) values, or a string giving a hex value (e.g. "#FF0000" for red).
        blink : bool
            Whether to blink the LED.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        ValueError
            if the color parameter is invalid or values are out of range
        """
        logger.info(
            f"Setting ColorLED '{self.name}' on channel {self.channel_id}: color={color}, blink={blink}"
        )
        try:
            self.client.set(self.channel_id, color, blink)
            logger.debug(
                f"ColorLED '{self.name}' successfully set to color={color}, blink={blink}"
            )
        except Exception as e:
            logger.error(
                f"Failed to set ColorLED '{self.name}' on channel {self.channel_id}: {e}"
            )
            raise

    def __repr__(self):
        return f"{self.name}; Channel ID: {self.channel_id}"



import logging

# import time
from pathlib import Path

from ekfsm.system import System

logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

config = Path(__file__).parent / "zip2-ccu-only.yaml"
system = System(config)

cpu = system[0]

ccu = system.ccu

# get parameterset
params = ccu.management.get_parameterset()
print(f"Parameterset: {params}")

# Firmware download
# def progress_callback(offset):
#     print(f"Loading firmware: {offset}")

# with open("/home/klaus/fw-ccu-00-default.bin", "rb") as f:
#     firmware = f.read()
#     ccu.management.load_firmware(firmware, progress_callback)
# time.sleep(5)

# Identify firmware
# print(f"CCU firmware: {ccu.management.identify_firmware()}")

# Fan control
# while True:
#     cputemp = cpu.th.cputemp()
#     print(f"CPU temperature: {cputemp}")
#     ccu.fan.push_temperature(-1, cputemp)
#     time.sleep(1)
#     fan_status = ccu.fan.fan_status(0)
#     print(f"Fan status0: {fan_status}")
#     fan_status = ccu.fan.fan_status(1)
#     print(f"Fan status1: {fan_status}")

# CCU Restart
# ccu.management.restart()

# # WD Trigger
# while True:
#     ccu.sysstate.wd_trigger()
#     time.sleep(5)



#!/usr/bin/env python3
"""License compliance analysis script for ekfsm project"""
import json
import yaml
import sys
from pathlib import Path
from datetime import datetime


def load_license_policy():
    """Load license policy configuration"""
    policy_file = Path(".license-policy.yaml")
    if not policy_file.exists():
        print("⚠️  No .license-policy.yaml found, using default policy", file=sys.stderr)
        return {
            "permitted_licenses": [
                "MIT",
                "MIT License",
                "BSD",
                "BSD License",
                "Apache-2.0",
                "Apache Software License",
                "ISC",
            ],
            "review_required": [
                "GPL",
                "LGPL",
                "MPL-2.0",
                "LGPL-2.1-or-later",
                "GNU Lesser General Public License",
            ],
            "prohibited_licenses": ["AGPL-1.0"],
            "package_exceptions": {},
            "compliance_settings": {
                "fail_on_prohibited": True,
                "warn_on_review_required": True,
                "fail_on_unknown": False,
                "require_license_files": True,
            },
        }

    try:
        with open(policy_file, "r") as f:
            return yaml.safe_load(f)
    except Exception as e:
        print(f"❌ Error loading license policy: {e}", file=sys.stderr)
        return None


def analyze_licenses():
    """Analyze licenses against policy"""
    try:
        policy = load_license_policy()
        if not policy:
            return 1

        with open("license-report.json", "r") as f:
            licenses = json.load(f)

        total_packages = len(licenses)
        permitted_count = 0
        review_required_count = 0
        prohibited_count = 0
        unknown_count = 0

        compliance_report = {
            "scan_timestamp": datetime.now().isoformat(),
            "total_packages": total_packages,
            "license_breakdown": {},
            "compliance_summary": {},
            "potential_issues": [],
            "policy_violations": [],
        }

        exit_code = 0

        for pkg in licenses:
            pkg_name = pkg["Name"]
            license_name = pkg.get("License", "Unknown").strip()

            # Check for package-specific overrides
            if pkg_name in policy.get("package_exceptions", {}):
                license_name = policy["package_exceptions"][pkg_name]

            # Count licenses
            if license_name in compliance_report["license_breakdown"]:
                compliance_report["license_breakdown"][license_name] += 1
            else:
                compliance_report["license_breakdown"][license_name] = 1

            # Categorize license
            is_permitted = any(
                perm in license_name for perm in policy.get("permitted_licenses", [])
            )
            is_review_required = any(
                rev in license_name for rev in policy.get("review_required", [])
            )
            is_prohibited = any(
                pro in license_name for pro in policy.get("prohibited_licenses", [])
            )
            is_unknown = (
                license_name in ["Unknown", "UNKNOWN"] or not license_name.strip()
            )

            if is_prohibited:
                prohibited_count += 1
                compliance_report["policy_violations"].append(
                    {
                        "package": pkg_name,
                        "version": pkg["Version"],
                        "license": license_name,
                        "violation_type": "PROHIBITED",
                        "severity": "HIGH",
                    }
                )
                if policy.get("compliance_settings", {}).get(
                    "fail_on_prohibited", True
                ):
                    exit_code = 1

            elif is_review_required:
                review_required_count += 1
                compliance_report["potential_issues"].append(
                    {
                        "package": pkg_name,
                        "version": pkg["Version"],
                        "license": license_name,
                        "issue": "Copyleft license - requires legal review",
                    }
                )
                if policy.get("compliance_settings", {}).get(
                    "warn_on_review_required", True
                ):
                    print(f"⚠️  {pkg_name}: {license_name} requires review")

            elif is_unknown:
                unknown_count += 1
                compliance_report["potential_issues"].append(
                    {
                        "package": pkg_name,
                        "version": pkg["Version"],
                        "license": license_name,
                        "issue": "Unknown license - manual review required",
                    }
                )
                if policy.get("compliance_settings", {}).get("fail_on_unknown", False):
                    exit_code = 1

            elif is_permitted:
                permitted_count += 1
            else:
                # License not explicitly categorized
                compliance_report["potential_issues"].append(
                    {
                        "package": pkg_name,
                        "version": pkg["Version"],
                        "license": license_name,
                        "issue": "License not in policy - requires review",
                    }
                )

        risk_level = (
            "HIGH"
            if (prohibited_count > 0)
            else (
                "MEDIUM" if (review_required_count > 0 or unknown_count > 0) else "LOW"
            )
        )

        compliance_report["compliance_summary"] = {
            "permitted_licenses": permitted_count,
            "review_required_licenses": review_required_count,
            "prohibited_licenses": prohibited_count,
            "unknown_licenses": unknown_count,
            "risk_level": risk_level,
            "policy_compliant": len(compliance_report["policy_violations"]) == 0,
        }

        with open("license-compliance-report.json", "w") as f:
            json.dump(compliance_report, f, indent=2)

        # Print summary to console
        print("📊 License Compliance Summary:")
        print(f"  Total packages: {total_packages}")
        print(f"  Permitted licenses: {permitted_count}")
        print(f"  Require review: {review_required_count}")
        print(f"  Prohibited licenses: {prohibited_count}")
        print(f"  Unknown licenses: {unknown_count}")
        print(f"  Risk level: {risk_level}")

        if compliance_report["policy_violations"]:
            print(
                f'❌ Found {len(compliance_report["policy_violations"])} policy violations'
            )
            for violation in compliance_report["policy_violations"]:
                print(
                    f'   - {violation["package"]} ({violation["license"]}) - {violation["violation_type"]}'
                )

        if (
            compliance_report["potential_issues"]
            and not compliance_report["policy_violations"]
        ):
            print(
                f'⚠️  Found {len(compliance_report["potential_issues"])} items requiring review'
            )

        if exit_code == 0 and not compliance_report["potential_issues"]:
            print("✅ All licenses compliant with policy")

        return exit_code

    except Exception as e:
        print(f"❌ Error analyzing licenses: {e}", file=sys.stderr)
        return 1


if __name__ == "__main__":
    exit_code = analyze_licenses()
    sys.exit(exit_code)



from pathlib import Path

from ekfsm.core.components import HWModule
from ekfsm.core.sysfs import SysfsDevice, sysfs_root
from ekfsm.log import ekfsm_logger

from .generic import Device

logger = ekfsm_logger(__name__)


class SMBIOS(Device):
    """
    A class to represent the SMBIOS device.

    A SMBIOS device is a virtual device that is used to read system
    configuration values from the DMI table.

    Note:
    Currently, only the board version / revision is read from the DMI table.
    """

    def __init__(
        self,
        name: str,
        parent: HWModule | None = None,
        children: list["Device"] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        logger.debug(f"Initializing SMBIOS device '{name}'")

        try:
            dmi_path = sysfs_root() / Path("devices/virtual/dmi/id")
            self.sysfs_device: SysfsDevice | None = SysfsDevice(dmi_path, False)
            logger.info(f"SMBIOS '{name}' initialized with DMI table at {dmi_path}")
        except Exception as e:
            logger.error(f"Failed to initialize SMBIOS '{name}' with DMI table: {e}")
            raise

        super().__init__(name, parent, None, abort, *args, **kwargs)

    def revision(self) -> str:
        """
        Get the board revision from the DMI table.

        Returns
        -------
        str
            The board revision.
        """
        logger.debug(f"Reading board revision for SMBIOS '{self.name}'")
        try:
            revision = self.sysfs.read_utf8("board_version")
            logger.debug(f"SMBIOS '{self.name}' board revision: {revision}")
            return revision
        except Exception as e:
            logger.error(f"Failed to read board revision for SMBIOS '{self.name}': {e}")
            raise



from abc import ABC, abstractmethod


class ProbeableDevice(ABC):
    @abstractmethod
    def probe(self, *args, **kwargs) -> bool:
        """
        Probe the hardware device to check if it is present and if it is the correct device.
        """
        pass



from .components import HWModule  # noqa: F401
from .probe import ProbeableDevice  # noqa: F401
from .slots import Slot, Slots, SlotType  # noqa: F401
from .sysfs import SysFSAttribute, SysfsDevice  # noqa: F401

__all__ = [
    "Slot",
    "SlotType",
    "Slots",
    "SysfsDevice",
    "SysFSAttribute",
    "HWModule",
    "ProbeableDevice",
]



import munch

from ekfsm.config import load_config


def test_config():
    config = load_config("tests/data/cfg_simple.yaml")
    assert isinstance(config, munch.Munch)

    assert config.system_config.name == "Simple System"
    assert len(config.system_config.slots) == 2
    assert config.system_config.slots[0].name == "SLOT1"
    assert config.system_config.slots[0].slot_type == "CPCI_S0_SYS"
    assert config.system_config.slots[0].desired_hwmodule_type == "EKF SC9"
    assert config.system_config.slots[0].desired_hwmodule_name == "CPU1"
    assert not hasattr(config.system_config.slots[0], "attributes")

    assert config.system_config.slots[1].name == "SLOT7"
    assert config.system_config.slots[1].slot_type == "CPCI_S0_PER"
    assert config.system_config.slots[1].desired_hwmodule_type == "EKF SUR"
    assert config.system_config.slots[1].desired_hwmodule_name == "SER1"
    attr = config.system_config.slots[1].attributes
    assert attr.slot_coding == 0



import pprint
import logging
from pathlib import Path
import cProfile

# import pstats

from ekfsm.system import System

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)

# Initialize system first (without profiling the setup)
config = Path(__file__).parent / "cctv.yaml"
try:
    system = System(config, abort=True)
    system_initialized = True
except Exception as e:
    print(f"System initialization failed: {e}")
    print("Continuing with mock profiling demonstration...")
    system_initialized = False
    system = None

if not system_initialized:
    print("System not available, creating mock profiling demonstration...")
    print(
        "This demonstrates the profiling structure that would be used with real hardware."
    )
    print("The same profiling patterns from sq3-only.py can be applied here.")
    exit(0)

# Start profiling HIER - nach System-Setup, vor io4edge_client Aufrufen
profiler = cProfile.Profile(builtins=False)
profiler.enable()

try:
    pprint.pprint(f"System slots {system.slots}")

    system.print()

    cpu = system["CPU"]
    cpuB = system.cpu
    cpuC = system[0]

    assert cpu == cpuB == cpuC

    # To check why below is failing
    # cpu_slot = system.slots["SYSTEM_SLOT"]
    # cpu_slotB = system.slots.SYSTEM_SLOT
    # cpu_slotC = system.slots[0]

    # assert cpu_slot == cpu_slotB == cpu_slotC

    cpu.print()
    print(f"probing CPU: {cpu.probe()}")
    print(
        f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}"
    )

    smc = system.smc

    smc.print()

    i4e = smc.i4e
    i4e.watchdog.kick()
    i4e.leds.led2.set(0, True)
    i4e.leds.led2.get()
    i4e.leds.led5.set(3, True)
    i4e.leds.led5.get()
    i4e.leds.led3.set(5, False)
    i4e.leds.led3.get()

    # button_array = i4e.buttons

    # eject = button_array.eject

    # eject.handler = lambda: print("Eject pressed")

    # stop_event = threading.Event()
    # button_thread = threading.Thread(target=button_array.read, args=(stop_event,))
    # button_thread.start()

    # for i in range(30):
    #     print("Main thread running...")
    #     i4e.watchdog.kick()
    #     i4e.leds.led2.set(0, True)
    #     i4e.leds.led5.set(3, True)
    #     i4e.leds.led3.set(5, False)
    #     sleep(1)

    # # To stop the thread:
    # stop_event.set()
    # button_thread.join()

except Exception as e:
    print(f"Error during execution: {e}")
    print("This is expected on systems without the required hardware.")

finally:
    profiler.disable()

    # print("\n" + "=" * 80)
    # print("PROFILING RESULTS - ekfsm and io4edge_client modules (CCTV)")
    # print("=" * 80)

    # # Create stats object and filter for our modules
    # stats = pstats.Stats(profiler)
    # stats.strip_dirs()  # clean paths

    # # Print comprehensive profiling results
    # print("\n1. TOP FUNCTIONS BY TOTAL TIME (including subcalls)")
    # stats.sort_stats("tottime").print_stats()

    # print("\n2. TOP FUNCTIONS BY CUMULATIVE TIME (time + subcalls)")
    # stats.sort_stats("cumtime").print_stats()

    # print("\n3. TOP FUNCTIONS BY NUMBER OF CALLS")
    # stats.sort_stats("ncalls").print_stats()

    # print("\n3a. ALL FUNCTIONS - DEBUGGING VIEW (to see what's actually being called)")
    # stats.sort_stats("tottime").print_stats(50)  # Show top 50 functions overall

    # print("\n4. CALLERS/CALLEES ANALYSIS FOR TOP 5 FUNCTIONS")
    # # Get the top 5 functions by total time
    # top_funcs = stats.get_stats_profile().func_profiles
    # filtered_funcs = {
    #     k: v
    #     for k, v in top_funcs.items()
    #     if any(mod in str(k) for mod in ["ekfsm", "io4edge_client"])
    # }
    # sorted_funcs = sorted(
    #     filtered_funcs.items(), key=lambda x: x[1].tottime, reverse=True
    # )

    # for i, (func_key, func_stats) in enumerate(sorted_funcs[:5]):
    #     print(f"\n--- Function #{i+1}: {func_key} ---")
    #     print(f"Total time: {func_stats.tottime:.6f}s, Calls: {func_stats.ncalls}")
    #     stats.print_callers(func_key)

    # # Save detailed profile to file for later analysis
    # profile_file = Path(__file__).parent / "cctv_profile.prof"
    # profiler.dump_stats(str(profile_file))
    # print(f"\n5. Detailed profile saved to: {profile_file}")
    # print("   Use: python -m pstats cctv_profile.prof")
    # print("   Or: snakeviz cctv_profile.prof (install: pip install snakeviz)")

    # print("\n" + "=" * 80)



from pathlib import Path
from typing import Generator

_CFG_DIR = Path(__file__).parent / "boards"


def find_board_config(module_type: str) -> Path | None:
    """
    Find a matching board config in `boards/oem/` given the module type specified in
    the system configuration file.

    Parameters
    ----------
    module_type
        Board type specified in the system configuration for a slot.
        It must consist of an OEM and the board type, separated by whitespace. Neither
        part may contain any other whitespace.
    """
    oem, board = module_type.split(maxsplit=1)
    if (
        path := _CFG_DIR / "oem" / oem.strip().lower() / f"{board.strip().lower()}.yaml"
    ).exists():
        return path
    return None


def all_board_cfg_files() -> Generator[Path, None, None]:
    r"""
    Generator that recursively yields all \*.yaml files in the config directory.

    Yields
    ------
    item: :class:`~pathlib.Path`
        Path to a config file.
    """
    path = Path(_CFG_DIR)
    for item in path.rglob("*.yaml"):
        if item.is_file():
            yield item


def next_or_raise(it, exc):
    value = next(it, None)
    if value is None:
        raise exc
    return value



from ekfsm.devices.coretemp import CoreTemp
from ekfsm.devices.generic import Device
from ekfsm.devices.smbios import SMBIOS
from ekfsm.devices.thermal_humidity import ThermalHumidity

from .eeprom import EEPROM, EKF_CCU_EEPROM, EKF_EEPROM
from .ekf_ccu_uc import EKFCcuUc
from .ekf_sur_led import EKFSurLed
from .gpio import GPIO, EKFIdentificationIOExpander, GPIOExpander
from .iio_thermal_humidity import IIOThermalHumidity
from .mux import I2CMux, MuxChannel
from .pmbus import PMBus, PSUStatus
from .io4edge import IO4Edge, GPIOArray
from .pixelDisplay import PixelDisplay
from .buttons import ButtonArray
from .buttons import Button
from .leds import LEDArray, ColorLED
from .toggles import BinaryToggle
from .ssm import SSM

CLASS_MAP = {
    "GenericDevice": Device,
    "I2CMux": I2CMux,
    "MuxChannel": MuxChannel,
    "GPIO": GPIO,
    "GPIOExpander": GPIOExpander,
    "EKFIdentificationIOExpander": EKFIdentificationIOExpander,
    "EEPROM": EEPROM,
    "EKF_EEPROM": EKF_EEPROM,
    "EKF_CCU_EEPROM": EKF_CCU_EEPROM,
    "EKFCcuUc": EKFCcuUc,
    "PMBus": PMBus,
    "PSUStatus": PSUStatus,
    "SMBIOS": SMBIOS,
    "HWMON": CoreTemp,
    "EKFSurLed": EKFSurLed,
    "IIOThermalHumidity": IIOThermalHumidity,
    "ThermalHumidity": ThermalHumidity,
    "IO4Edge": IO4Edge,
    "PixelDisplay": PixelDisplay,
    "ButtonArray": ButtonArray,
    "Button": Button,
    "ColorLED": ColorLED,
    "LEDArray": LEDArray,
    "GPIOArray": GPIOArray,
    "BinaryToggle": BinaryToggle,
    "SSM": SSM,
}



from __future__ import annotations

import logging
from pprint import pformat
from typing import TYPE_CHECKING, Any, List

from schema import Optional, Or, Schema, SchemaError, Use
from termcolor import colored

from ekfsm.devices import CLASS_MAP
from ekfsm.exceptions import ConfigError

if TYPE_CHECKING:
    from ekfsm.devices.generic import Device

    from .components import HWModule


def import_board(logger: logging.Logger, data, parent=None, abort: bool = False):
    from .components import HWModule

    device_type = data.get("device_type")
    nodecls = CLASS_MAP.get(device_type)
    if nodecls is None:
        raise ConfigError(f"Unknown device type: {device_type}")

    children = data.pop("children", [])
    if parent is not None and isinstance(parent, HWModule):
        # ???
        pass

    if provides := data.get("provides"):
        for p in provides:
            interfaces = data["provides"][p]

            for interface in interfaces:
                if isinstance(interface, str):
                    if attr := getattr(nodecls, interface, None):
                        if not callable(attr) and not abort:
                            raise ConfigError("No such method")
                elif isinstance(interface, dict):
                    for key, value in interface.items():
                        if attr := getattr(nodecls, value, None):
                            if not callable(attr) and not abort:
                                raise ConfigError("No such method")
                else:
                    raise ConfigError("Error in board configuration")

    node = nodecls(parent=parent, abort=abort, **data)

    if children is not None:
        for child in children:
            try:
                logger.debug(f"Importing sub device {pformat(child)}")
                import_board(logger, data=child, parent=node, abort=abort)
            except Exception as e:
                if abort:
                    logger.error(
                        f"Failed to import sub device {pformat(child)}: {e}. Aborting."
                    )
                    raise e
                else:
                    logger.error(
                        f"Failed to import sub device {pformat(child)}: {e}. Continuing anyway."
                    )
    return node


def provides_validator(x: Any) -> Any:
    if isinstance(x, str):
        return x
    elif isinstance(x, dict) and len(x) == 1:
        key, value = next(iter(x.items()))
        if isinstance(key, str) and isinstance(value, str):
            return x
    raise SchemaError(
        "Each provides item must be either a string or a dictionary with one string key/value pair"
    )


device_schema = Schema({})

_device_structure = Schema(
    {
        "device_type": str,
        "name": str,
        Optional("addr"): int,
        Optional("slot_coding_mask"): int,
        Optional("channel_id"): int,
        Optional("service_suffix"): str,
        Optional("keepaliveInterval"): int,
        Optional("provides"): {
            Optional(str): [Use(provides_validator)],
        },
        Optional("children"): Or(None, [device_schema]),
    }
)

device_schema._schema = _device_structure

module_schema = Schema(
    {
        "id": Or(int, str),
        "name": str,
        "slot_type": str,
        Optional("children"): [device_schema],
        Optional("bus_masters"): {
            Optional("i2c"): dict,
        },
    }
)


def deserialize_module(logger: logging.Logger, data: dict) -> tuple[str, str, str]:
    """
    docstring
    """
    module_schema.validate(data)

    id, name, slot_type = (data[key] for key in ["id", "name", "slot_type"])
    logger.debug(colored(f"Importing top level module {pformat(name)}", "green"))

    return id, name, slot_type


def deserialize_hardware_tree(
    logger: logging.Logger, data: dict, parent: "HWModule"
) -> List["Device"]:
    abort = parent.abort

    module_schema.validate(data)

    children = data.pop("children", None)
    if not children:
        return []

    devices = []
    for child in children:
        try:
            logger.debug(
                colored(f"Importing top level device {pformat(child)}", "green")
            )

            node = import_board(logger, child, parent=parent, abort=abort)
            devices.append(node)
        except Exception as e:
            logger.error(
                colored(
                    f"Failed to import top-level device {pformat(child)}: {e}", "red"
                )
            )
            logger.error(colored("Aborting." if abort else "Continuing anyway", "red"))
            if abort:
                raise

    return devices



import threading
import pprint
import logging

# from time import sleep
from pathlib import Path
from ekfsm.system import System


logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)

# Initialize system first (without profiling the setup)
config = Path(__file__).parent / "cctv.yaml"
system = System(config, abort=True)

pprint.pprint(f"System slots {system.slots}")

system.print()

cpu = system["CPU"]
cpuB = system.cpu
cpuC = system[0]

assert cpu == cpuB == cpuC

# To check why below is failing
# cpu_slot = system.slots["SYSTEM_SLOT"]
# cpu_slotB = system.slots.SYSTEM_SLOT
# cpu_slotC = system.slots[0]

# assert cpu_slot == cpu_slotB == cpu_slotC

cpu.print()
print(f"probing CPU: {cpu.probe()}")
print(
    f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}"
)

smc = system.smc

smc.print()

# Test 4 connection in parallels to io4edge
i4e = smc.i4e

# Test watchdog client connection
i4e.watchdog.client.open()

# open led client connection
i4e.leds.client.open()

# open power and redriver client connections
pw_toggle = i4e.gpios.power
redriver_toggle = i4e.gpios.redriver

i4e.gpios.client.open()

# open button client connection
button_array = i4e.gpios.buttons

eject = button_array.eject

eject.handler = lambda: print("Eject pressed")

stop_event = threading.Event()
button_thread = threading.Thread(target=button_array.read, args=(stop_event,))


# operate
button_thread.start()

for i in range(5):
    print("Main thread running...")
    i4e.watchdog.kick()

    # Test LED operations
    i4e.leds.led2.set(0, True)
    led2 = i4e.leds.led2.get()
    assert led2 == (0, True)
    i4e.leds.led5.set(3, True)
    led5 = i4e.leds.led5.get()
    assert led5 == (3, True)
    i4e.leds.led3.set(5, False)
    led3 = i4e.leds.led3.get()
    assert led3 == (5, False)

    pw_toggle.on()
    # sleep(0.5)
    pw_toggle.off()
    # sleep(0.5)
    redriver_toggle.on()
    # sleep(0.5)
    redriver_toggle.off()
    # sleep(0.5)

i4e.leds.client.close()
i4e.gpios.client.close()
i4e.watchdog.client.close()

# To stop the thread:
stop_event.set()
button_thread.join()



import logging
from pathlib import Path

import pytest

from ekfsm.simctrl import enable_simulation, register_gpio_simulations
from ekfsm.system import System


@pytest.fixture
def simulation():
    enable_simulation(Path(__file__).parent / "sim" / "sys")
    register_gpio_simulations()


def test_system(simulation):
    config = Path(__file__).parent / "sim" / "test_system.yaml"
    system = System(config, abort=True)

    ccu = system.ccu
    assert ccu is not None

    eeprom = ccu.mux.ch00.eeprom

    eeprom.write_cmodel("SRS-C001")
    assert eeprom.cmodel() == "SRS-C001"

    eeprom.write_cvendor("EKF Elektronik")
    assert eeprom.cvendor() == "EKF Elektronik"

    eeprom.write_crevision("2.0")
    assert eeprom.crevision() == "2.0"

    eeprom.write_unit(1)
    assert eeprom.unit() == 1

    data = b"\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10"
    ccu.custom_eeprom.write(data)
    assert ccu.custom_eeprom.read()[0 : len(data)] == data


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    pytest.main([__file__])



import logging
from pathlib import Path

from ekfsm.system import HWModule, System

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")


def print_inventory(name, inventory):
    print(f"{name}: {inventory.vendor()} {inventory.model()} {inventory.serial()}")


def probe_and_print_inventory(board: HWModule):
    board.probe()
    print_inventory(board.name, board.inventory)


config = Path(__file__).parent / "zip2-comp1.yaml"
system = System(config, abort=False)
print(system)

_psu = system.psu
probe_and_print_inventory(_psu)

_cpu = system.cpu
probe_and_print_inventory(_cpu)

_lan = system.lan
probe_and_print_inventory(_lan)

_ssd1 = system.ssd1
probe_and_print_inventory(_ssd1)

_kom = system.kom
probe_and_print_inventory(_kom)

_io = system.io
board = _io
board.probe()
print(f"inventory: {board.inventory.vendor()} {board.inventory.model()} ")

probe_and_print_inventory(_io)

_ssd2 = system.ssd2
probe_and_print_inventory(_ssd2)

_ser = system.ser
probe_and_print_inventory(_ser)

_ccu = system.ccu
probe_and_print_inventory(_ccu)


# SUR LED Test
_ser.led_a.set(0, "red")
_ser.led_a.set(1, "yellow")
_ser.led_c.set(0, "white")
_ser.led_d.set(1, "blue")

print(f"SUR TEMP: {_ser.th.temperature()}")
print(f"SUR HUMIDITY: {_ser.th.humidity()}")


# CCU Test
# chassis inventory via ccu object
print_inventory("chassis", _ccu.chassis_inventory)
# alternative way to get chassis inventory
print_inventory("chassis", system.inventory)

print(f"CCU firmware: {_ccu.management.identify_firmware()}")

# SPV Test
print(f"spv gpio0 control lines {_kom.gpio0.num_lines()}")
print(f"spv gpio1 control lines {_kom.gpio1.num_lines()}")
print(f"spv gpio2 control lines {_kom.gpio2.num_lines()}")



import subprocess
import sys
import time

import pytest

# start 2 times test_lock.py and wait for them to finish


def test_locking():
    """
    Test the locking mechanism by running two instances of the same script in parallel.
    """

    # Start the first instance
    proc1 = subprocess.Popen([sys.executable, "tests/locking/lock_tester.py", "1"])
    time.sleep(1)  # Give it some time to start

    # Start the second instance
    proc2 = subprocess.Popen([sys.executable, "tests/locking/lock_tester.py", "2"])

    # Wait for both processes to finish
    proc1.wait()
    proc2.wait()

    assert proc1.returncode == 0
    assert proc2.returncode == 0


if __name__ == "__main__":
    pytest.main([__file__])



"""
Some devices or device functions don't allow concurrent access.
The locking mechanism is used to ensure that only one process/thread can
access the device (function) at a time.

Locking granularity is defined by the device.
It may be at the device level or function level.

Application can choose to use the locking mechanism or not.
By default, the locking mechanism is enabled and uses the default
lockfile root directory ``/var/lock/ekfsm``.

Use :func:`locking_configure` to enable or disable the locking mechanism or to change
the lockfile root directory.
"""

import fcntl
import os
from contextlib import contextmanager
from pathlib import Path
from typing import List

USE_LOCK = True
LOCKFILE_ROOT = "/var/lock/ekfsm"
ALL_LOCKERS: List["Locker"] = []  # List of all locker instances


def locking_configure(enable: bool, lockfile_root: str = LOCKFILE_ROOT):
    """
    Configures the locking mechanism.

    Parameters
    ----------
    enable
        Whether to enable or disable locking.
    lockfile_root
        The root directory for lockfiles.
    """
    global USE_LOCK, LOCKFILE_ROOT
    USE_LOCK = enable
    LOCKFILE_ROOT = lockfile_root


def locking_cleanup():
    """
    Cleans up all lockfiles and closes all lock file descriptors.
    Should be called at the end of the program to ensure all locks are released.
    """
    for locker in ALL_LOCKERS:
        locker.cleanup()


class Locker:
    """
    A class that implements a locking mechanism using file locks.

    Parameters
    ----------
    module
        The name of the module or resource to lock. This will be used to create a unique lock file.

    Example
    -------
    .. code-block:: python

        with Locker("mysharedresourcename").lock():
            # Access the shared resource here
            pass

    """

    def __init__(self, module: str):
        if not USE_LOCK:
            return

        self.lockfile_path = Path(LOCKFILE_ROOT) / module
        self.lock_fd = None
        # Ensure lockfile exists
        os.makedirs(LOCKFILE_ROOT, exist_ok=True)
        open(self.lockfile_path, "a").close()
        ALL_LOCKERS.append(self)

    def cleanup(self):
        """
        Cleans up the lock file and closes the lock file descriptor.

        Important
        ---------
        This method should be called when the lock is no longer needed.

        Note
        ----
        It is automatically called when the context manager exits.
        """
        if not USE_LOCK:
            return
        if self.lock_fd is not None:
            try:
                fcntl.flock(self.lock_fd, fcntl.LOCK_UN)
                os.close(self.lock_fd)
                self.lock_fd = None
            except (OSError, AttributeError):
                pass

    @contextmanager
    def lock(self):
        """
        Locks the resource for exclusive access.

        Note
        ----
        This method is a context manager that locks the resource when entered
        and releases the lock when exited.
        """
        if not USE_LOCK:
            yield
        self.lock_fd = os.open(self.lockfile_path, os.O_RDWR)
        try:
            fcntl.flock(self.lock_fd, fcntl.LOCK_EX)
            yield
        finally:
            self.cleanup()



#!/usr/bin/env python3
"""
Template für Profiling von ekfsm und io4edge_client Modulen

Dieses Template zeigt, wie das Profiling-Prinzip auf beliebige Scripts angewendet wird.
Kopieren Sie diesen Code und passen Sie ihn an Ihr spezifisches Script an.

Verwendung:
1. Importieren Sie cProfile und pstats
2. Initialisieren Sie das System OHNE Profiling
3. Starten Sie das Profiling VOR den io4edge_client Aufrufen
4. Führen Sie Ihre io4edge_client/ekfsm Operationen aus
5. Stoppen Sie das Profiling und analysieren Sie die Ergebnisse
"""

import cProfile
import pstats
from pathlib import Path
import logging

# Optional: Setzen Sie Logging-Level für bessere Sichtbarkeit
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)


def profile_ekfsm_io4edge_operations():
    """
    Template-Funktion für Profiling von ekfsm und io4edge_client Operationen
    """

    # SCHRITT 1: System-Initialisierung OHNE Profiling
    print("Initializing system...")
    try:
        from ekfsm.system import System

        # Passen Sie den Konfigurationspfad an Ihr Script an
        config = Path(__file__).parent / "your-config.yaml"
        system = System(config, abort=True)
        system_available = True

    except Exception as e:
        print(f"System initialization failed: {e}")
        print("Continuing with demonstration of profiling structure...")
        system_available = False

    if not system_available:
        print("\n" + "=" * 60)
        print("PROFILING TEMPLATE STRUCTURE")
        print("=" * 60)
        print("1. Import cProfile and pstats")
        print("2. Initialize system WITHOUT profiling")
        print("3. Start profiler BEFORE io4edge_client calls")
        print("4. Execute your io4edge_client operations")
        print("5. Stop profiler and analyze results")
        print("6. Filter results for ekfsm|io4edge_client modules")
        return

    # SCHRITT 2: Profiling starten - NACH System-Setup, VOR io4edge_client Aufrufen
    print("Starting profiler...")
    profiler = cProfile.Profile(builtins=False)
    profiler.enable()

    try:
        # SCHRITT 3: Ihre io4edge_client/ekfsm Operationen hier
        print("Executing monitored operations...")

        # Beispiel-Operationen (passen Sie diese an Ihr Script an):

        # System-Operationen
        system.print()

        # Beispiel: CPU-Zugriff
        cpu = system["CPU"]  # Ändern Sie "CPU" entsprechend Ihrer Konfiguration
        cpu.print()

        # Beispiel: Hardware-Module Zugriff
        # board = system["your_board_name"]
        # i4e = board.your_io4edge_device

        # Beispiel: io4edge_client Operationen
        # i4e.watchdog.kick()
        # i4e.leds.led1.set(0, True)
        # button_array = i4e.buttons
        # button_array.some_button.handler = lambda: print("Button pressed")

        # Beispiel: Streaming-Operationen
        # stop_event = threading.Event()
        # button_thread = threading.Thread(target=button_array.read, args=(stop_event,))
        # button_thread.start()
        # time.sleep(5)  # Simulate some work
        # stop_event.set()
        # button_thread.join()

        print("Operations completed successfully")

    except Exception as e:
        print(f"Error during monitored operations: {e}")

    finally:
        # SCHRITT 4: Profiling stoppen und analysieren
        profiler.disable()

        print("\n" + "=" * 80)
        print("PROFILING RESULTS - ekfsm and io4edge_client modules")
        print("=" * 80)

        # Stats-Objekt erstellen
        stats = pstats.Stats(profiler)
        stats.strip_dirs()

        # SCHRITT 5: Analysieren Sie die Ergebnisse

        print("\n1. TOP FUNCTIONS BY TOTAL TIME")
        # Passen Sie den Regex-Filter an Ihre spezifischen Module an
        stats.sort_stats("tottime").print_stats(r"ekfsm|io4edge_client|your_module", 20)

        print("\n2. TOP FUNCTIONS BY CUMULATIVE TIME")
        stats.sort_stats("cumtime").print_stats(r"ekfsm|io4edge_client|your_module", 20)

        print("\n3. TOP FUNCTIONS BY NUMBER OF CALLS")
        stats.sort_stats("ncalls").print_stats(r"ekfsm|io4edge_client|your_module", 15)

        print("\n4. ALL FUNCTIONS OVERVIEW")
        stats.sort_stats("tottime").print_stats(30)

        # SCHRITT 6: Profiling-Datei speichern für detaillierte Analyse
        profile_file = Path(__file__).parent / "your_script_profile.prof"
        profiler.dump_stats(str(profile_file))

        print(f"\n5. PROFILE FILE SAVED: {profile_file}")
        print(f"   Detailed analysis: python -m pstats {profile_file}")
        print(f"   Visual analysis: pip install snakeviz && snakeviz {profile_file}")

        print("\n" + "=" * 80)


def create_profile_analysis_script(script_name):
    """
    Erstellt ein Analyse-Script für ein spezifisches Profiling
    """
    analysis_script = f"""#!/usr/bin/env python3
'''
Analysis script for {script_name}_profile.prof
'''

import pstats
from pathlib import Path

def analyze_{script_name}_profile():
    profile_file = Path(__file__).parent / "{script_name}_profile.prof"

    if not profile_file.exists():
        print(f"Profile file not found: {{profile_file}}")
        return

    stats = pstats.Stats(str(profile_file))
    stats.strip_dirs()

    print("="*80)
    print("DETAILED ANALYSIS OF {script_name.upper()} PROFILING")
    print("="*80)

    # Modulspezifische Analysen
    print("\\n1. EKFSM MODULE ANALYSIS")
    stats.sort_stats("tottime").print_stats(r"ekfsm", 15)

    print("\\n2. IO4EDGE_CLIENT MODULE ANALYSIS")
    stats.sort_stats("tottime").print_stats(r"io4edge_client", 15)

    # Fügen Sie hier weitere spezifische Analysen ein

if __name__ == "__main__":
    analyze_{script_name}_profile()
"""

    analysis_file = Path(__file__).parent / f"analyze_{script_name}_profile.py"
    analysis_file.write_text(analysis_script)
    print(f"Analysis script created: {analysis_file}")


if __name__ == "__main__":
    print("PROFILING TEMPLATE FOR EKFSM AND IO4EDGE_CLIENT")
    print("=" * 50)
    profile_ekfsm_io4edge_operations()

    # Optional: Erstellen Sie ein Analyse-Script
    # create_profile_analysis_script("your_script_name")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

project = "ekfsm"
copyright = "2024, Klaus Popp, Felix Päßler, Jan Jansen"
author = "Klaus Popp, Felix Päßler, Jan Jansen"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",  # Auto-generate documentation from docstrings
    "sphinx.ext.autosummary",  # Generate neat summary tables
    "sphinx.ext.napoleon",  # Support for Google style docstrings
    "sphinx.ext.viewcode",  # Add links to source code
    "sphinx_autodoc_typehints",  # Auto-document type hints (optional)
    "sphinx.ext.coverage",
    "sphinx_rtd_theme",
    "sphinx.ext.intersphinx",
    "sphinx_click",
]

autodoc_typehints = "description"

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]

coverage_show_missing_items = True  # Shows missing docstrings in the output
coverage_write_headline = True  # Adds a headline to the coverage report

# Automatically generate summary pages
autosummary_generate = True

html_theme = "sphinx_rtd_theme"  # or another theme of your choice

# -- General settings ------------------------------------------
master_doc = "index"  # The master toctree document

# -- Autodoc options (optional) --------------------------------
autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "private-members": False,
    "show-inheritance": True,
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "anytree": ("https://anytree.readthedocs.io/en/latest/", None),
}

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_private_with_doc = True
napoleon_preprocess_types = True
napoleon_include_init_with_doc = True

# abort on error
nitpicky = True



import logging
import pprint
from pathlib import Path

import pytest

from ekfsm.devices.pmbus import PSUStatus
from ekfsm.simctrl import enable_simulation, register_gpio_simulations
from ekfsm.system import HWModule, System


@pytest.fixture
def simulation():
    enable_simulation(Path(__file__).parent / "sim" / "sys")
    register_gpio_simulations()


def test_system(simulation):
    config = Path(__file__).parent / "sim" / "test_system.yaml"
    system = System(config, abort=True)
    pprint.pprint(f"System slots {system.slots}")

    psu = system.psu
    assert psu is not None
    assert isinstance(psu, HWModule)
    assert psu.probe()
    assert psu.inventory.vendor() == "Hitron"
    assert psu.inventory.model() == "HDRC300S-110J-D120E(N)"
    assert psu.main.voltage() == 12.2
    assert psu.sby.current() == 0.5
    assert psu.th.temperature() == 22.0
    assert psu.main.status() == PSUStatus.OK
    assert psu.sby.status() == PSUStatus.OK

    cpu = system["CPU"]
    assert cpu is not None
    assert isinstance(cpu, HWModule)
    assert cpu.inventory.vendor() == "EKF Elektronik"
    assert cpu.inventory.model() == "SC9-TOCCATA"
    assert cpu.inventory.serial() == "53082029"

    srf = system.slots["SLOT1"].hwmodule
    assert srf is not None
    assert isinstance(srf, HWModule)

    assert srf.probe()

    assert srf.inventory.vendor() == "EKF Elektronik"
    assert srf.inventory.model() == "SRF-FAN"
    assert srf.inventory.revision() == "0"
    assert srf.inventory.serial() == "54021107"
    assert srf.mux.get_i2c_bus_number() == 1
    assert srf.mux.ch00.gpio.get_i2c_bus_number() == 9

    ccu = system.ccu
    assert ccu is not None
    version = ccu.mux.ch00.eeprom.version()
    assert version == 1


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    pytest.main([__file__])



from .system import System
from .config import load_config
from .log import ekfsm_logger
from .version import version

__all__ = ("System", "load_config", "ekfsm_logger", "version")

__version__ = version



from collections.abc import MutableMapping
from pathlib import Path
from typing import Callable, List, Union

from more_itertools import first_true

from ekfsm.exceptions import ConversionError, DriverError, SysFSError

SYSFS_ROOT = Path("/sys")


def sysfs_root() -> Path:
    return SYSFS_ROOT


def set_sysfs_root(path: Path) -> None:
    global SYSFS_ROOT
    SYSFS_ROOT = path


def file_is_sysfs_attr(path: Path) -> bool:
    return path.is_file() and not path.stat().st_mode & 0o111


class SysFSAttribute:
    """
    A SysFSAttribute is a singular sysfs attribute located somewhere in */sys*.

    Parameters
    ----------
    path: :class:`~pathlib.Path`
        Path to the underlying file for the SysFSAttribute instance.
    """

    def __init__(self, path: Path):
        if not path.exists() or not path.is_file():
            raise FileNotFoundError("Invalid sysfs attribute path")

        self.path = path
        self.name: str = path.name

    def read_utf8(self) -> str:
        try:
            return self.path.read_text()
        except OSError as e:
            raise SysFSError("Error accessing SysFS attribute") from e

    def read_bytes(self) -> bytes:
        try:
            return self.path.read_bytes()
        except OSError as e:
            raise SysFSError("Error accessing SysFS attribute") from e

    # FIXME: This cannot work due to sysfs attributes not supporting seek().
    def write(self, data: Union[str, bytes, None], offset: int = 0) -> None:
        if self.is_sysfs_attr() and data is not None:
            mode = "r+" if isinstance(data, str) else "rb+"
            try:
                with open(self.path, mode) as f:
                    f.seek(offset)
                    f.write(data)
            except OSError as e:
                raise SysFSError("Error accessing SysFS attribute") from e

    def is_sysfs_attr(self) -> bool:
        return file_is_sysfs_attr(self.path)

    def __repr__(self):
        return f"SysFSAttribute: {self.name}"


def list_sysfs_attributes(path: Path) -> List[SysFSAttribute]:
    if not path.exists() or not path.is_dir():
        raise FileNotFoundError(f"Invalid sysfs directory: {path}")

    return [SysFSAttribute(item) for item in path.iterdir() if file_is_sysfs_attr(item)]


class SysfsDevice(MutableMapping):
    def __init__(
        self,
        base_dir: Path,
        driver_required=True,
        find_driver: Callable | None = None,
    ):
        self.path: Path = base_dir
        self.driver_required = driver_required

        try:
            self.driver = self.get_driver()
        except Exception:
            self.driver = None

            if driver_required:
                raise DriverError(f"No driver found for device at {base_dir}")

        try:
            self.attributes: List[SysFSAttribute] = list_sysfs_attributes(self.path)
        except FileNotFoundError as e:
            raise SysFSError(f"SysFS entry for {base_dir} does not exist") from e

    def __getitem__(self, key):
        if (
            attr := first_true(self.attributes, pred=lambda a: a.name == key)
        ) is not None:
            return attr

        raise KeyError(f"'{key}' is not a valid sysfs attribute in {self.path}")

    def __setitem__(self, key, value):
        self[key].write(value)

    def __delitem__(self, key):
        del self.attributes[key]

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)

    def pre(self) -> None:
        pass

    def post(self) -> None:
        pass

    def write_attr(self, attr: str, data: str | bytes) -> None:
        next(x for x in self.attributes if x.name == attr).write(data)

    def write_attr_bytes(self, attr: str, data: str) -> None:
        # TODO: This
        pass

    def read_attr_utf8(self, attr: str) -> str:
        return next(x for x in self.attributes if x.name == attr).read_utf8()

    def read_float(self, attr: str) -> float:
        """
        Read a sysfs attribute as a floating-point number

        Parameters
        ----------
        attr: str
            The sysfs attribute to read

        Returns
        -------
        float
            The sysfs attribute as a floating-point number

        Raises
        ------
        SysFSError
            If the sysfs attribute does not exist
        ConversionError
            If the sysfs attribute could not be converted to a floating-point number
        """
        try:
            value = self.read_attr_utf8(attr)
            return float(value)
        except StopIteration as e:
            raise SysFSError(f"'{attr}' sysfs attribute does not exist") from e
        except SysFSError:
            raise
        except ValueError as e:
            raise ConversionError(
                "Failed to convert sysfs value to floating-point value"
            ) from e

    def read_int(self, attr) -> int:
        """
        Read a sysfs attribute stored as a string as an integer

        Parameters
        ----------
        attr: str
            The sysfs attribute to read

        Returns
        -------
        int
            The sysfs attribute as an integer

        Raises
        ------
        SysFSError
            If the sysfs attribute does not exist
        ConversionError
            If the sysfs attribute could not be converted to an integer
        """
        try:
            value = self.read_attr_utf8(attr).strip()
            return int(value)
        except StopIteration as e:
            raise SysFSError(f"'{attr}' sysfs attribute does not exist") from e
        except SysFSError:
            raise
        except ValueError as e:
            raise ConversionError("Failed to convert sysfs value to int") from e

    def read_hex(self, attr) -> int:
        """
        Read a sysfs attribute stored as a hexadecimal integer as integer

        Parameters
        ----------
        attr: str
            The sysfs attribute to read

        Returns
        -------
        int
            The sysfs attribute as a hexadecimal integer

        Raises
        ------
        SysFSError
            If the sysfs attribute does not exist
        ConversionError
            If the sysfs attribute could not be converted to a hexadecimal integer
        """
        try:
            value = self.read_attr_utf8(attr).strip()
            return int(value, 16)
        except StopIteration as e:
            raise SysFSError(f"'{attr}' sysfs attribute does not exist") from e
        except SysFSError:
            raise
        except ValueError as e:
            raise ConversionError("Failed to convert sysfs value to hex int") from e

    def read_utf8(self, attr, strip=True) -> str:
        """
        Read a sysfs attribute as a UTF-8 encoded string

        Parameters
        ----------
        attr: str
            The sysfs attribute to read
        strip: bool
            Strip whitespace, defaults to true

        Returns
        -------
        str
            The sysfs attribute as a UTF-8 encoded string

        Raises
        ------
        SysFSError
            If the sysfs attribute does not exist
        """
        try:
            value = self.read_attr_utf8(attr)
            if strip:
                value = value.strip()

            return value
        except StopIteration as e:
            raise SysFSError(f"'{attr}' sysfs attribute does not exist") from e
        except SysFSError:
            raise

    def read_attr_bytes(self, attr: str) -> bytes:
        return next(x for x in self.attributes if x.name == attr).read_bytes()

    def read_bytes(self, attr) -> bytes:
        try:
            value = self.read_attr_bytes(attr)
            return value
        except StopIteration as e:
            raise SysFSError(f"'{attr}' sysfs attribute does not exist") from e
        except SysFSError:
            raise

    def extend_attributes(self, attributes: List[SysFSAttribute]):
        self.attributes.extend(attributes)

    def get_driver(self) -> str | None:
        path = self.path

        if self.path.joinpath("device").exists():
            path = self.path.joinpath("device")
        elif not path.joinpath("driver").exists():
            raise DriverError("Failed to retrieve driver info")

        return path.joinpath("driver").readlink().name



from ekfsm.exceptions import ConversionError, SysFSError

from .generic import SysfsDevice


def iio_get_in_value(dev: SysfsDevice, attrset: str) -> float:
    """
    Calculate a value from an IIO in_* attribute set, using the _raw, _scale and _offset attributes (if present).
    Typical name of attrset are "in_temp" or "in_voltage0".

    Formula according to https://wiki.st.com/stm32mpu/wiki/How_to_use_the_IIO_user_space_interface

    Parameters
    ----------
    dev
        sysfs device object pointing to the iio directory
    attrset
        name of the attribute set to read from (e.g. "in_temp")

    Returns
    -------
    float
        calculated value from the attribute set (no unit conversion)

    Raises
    ------
    FileNotFoundError
        if the neiter _input nor _raw attribute is found

    """
    try:
        content = dev.read_float(f"{attrset}_input")
    except (SysFSError, ConversionError):
        try:
            raw = dev.read_float(f"{attrset}_raw")
            offset = dev.read_float(f"{attrset}_offset")
            scale = dev.read_float(f"{attrset}_scale")
            content = (raw + offset) * scale
        except (SysFSError, ConversionError) as e:
            raise FileNotFoundError from e

    return content



#!/usr/bin/env python3
"""
Enhanced profiling analysis script for cctv_profile.prof
Analyzes CCTV system profiling results focusing on:
- Watchdog operations
- LED control functions
- Button array operations
- ekfsm and io4edge_client module usage
"""

import pstats
from pathlib import Path


def analyze_cctv_profile():
    profile_file = Path(__file__).parent / "cctv_profile.prof"

    if not profile_file.exists():
        print(f"Profile file not found: {profile_file}")
        print("Run cctv.py first to generate the profile")
        return

    stats = pstats.Stats(str(profile_file))
    stats.strip_dirs()

    print("=" * 80)
    print("DETAILED ANALYSIS OF CCTV PROFILING")
    print("=" * 80)

    print("\n1. EKFSM MODULE ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"ekfsm", 15)

    print("\n2. IO4EDGE_CLIENT MODULE ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"io4edge_client", 15)

    print("\n3. WATCHDOG OPERATIONS ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"watchdog|kick", 10)

    print("\n4. LED CONTROL OPERATIONS ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"led|colorLED|set", 10)

    print("\n5. BUTTON OPERATIONS ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"button|eject|read_stream", 10)

    print("\n6. SOCKET/TRANSPORT LAYER ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"socket|transport", 10)

    print("\n7. MOST TIME-CONSUMING FUNCTIONS (ALL)")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(20)

    print("\n8. FUNCTIONS WITH MOST CALLS")
    print("-" * 40)
    stats.sort_stats("ncalls").print_stats(15)

    print("\n9. CUMULATIVE TIME ANALYSIS")
    print("-" * 40)
    stats.sort_stats("cumtime").print_stats(15)

    print("\n10. CCTV-SPECIFIC HARDWARE INTERACTIONS")
    print("-" * 40)
    # Look for CCTV-specific patterns
    stats.sort_stats("tottime").print_stats(r"smc|i4e", 10)


if __name__ == "__main__":
    analyze_cctv_profile()



from ekfsm.core.components import SysTree
from ekfsm.log import ekfsm_logger
from ekfsm.utils import next_or_raise

from ..core.sysfs import list_sysfs_attributes
from .generic import Device
from .iio import iio_get_in_value


class IIOThermalHumidity(Device):
    """
    Device for IIO thermal and/or humidity sensors.

    Parameters
    ----------
    name
        The name of the device.
    parent
        The parent device of the IIOThermalHumidity device. If None, no parent is created.
    children
        The children of the IIOThermalHumidity device. If None, no children are created.
    """

    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: list[Device] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        self.logger = ekfsm_logger("IIOThermalHumidity:" + name)
        super().__init__(name, parent, children, abort, *args, **kwargs)
        self.addr = self.get_i2c_chip_addr()
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr, False)

        # TODO: We can just search the attributes directly
        iio_dir = self.sysfs.path.glob("iio:device*")
        attrs = next_or_raise(iio_dir, FileNotFoundError("IIO entry not found"))

        self.sysfs_device.extend_attributes(list_sysfs_attributes(attrs))

    def temperature(self) -> float:
        return iio_get_in_value(self.sysfs, "in_temp") / 1000.0

    def humidity(self) -> float:
        return iio_get_in_value(self.sysfs, "in_humidityrelative") / 1000.0
